#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <float.h>
#include "tasks.h"
#include "resources.h"
#include "result.h"

	
/***
*** Dependent task placement and scheduling with function configuration in edge computing, IWQoS’19
*** 
***/

int eftf()
{
	memset(&rst, 0, sizeof(struct result));
	int i, j, k, l;
	double cur_dev_energy;
	float eft = FLT_MAX;
	int flag;//0 for dev, 1 for edge, 2 for cloud
	int eft_core = -1; //for dev or server
	struct dev *eft_dev = NULL;
	struct edge *eft_edge = NULL;//for edge
	struct server *eft_server = NULL;//for edge
	struct vm *eft_vm = NULL;//for cloud
	struct cloud *eft_cloud = NULL; // for new vm
	for(i=0; i<num_task; i++)
	{
		eft = FLT_MAX;
		flag = -1;
		struct dev *cur_dev = tasks[i].dev;
		//scheduling the task to the device
		for(j=0; j<cur_dev->core; j++)
		{
			cur_dev->use_time[j] += tasks[i].length / cur_dev->cap;
			cur_dev_energy = cal_energy_dev(cur_dev, 0.0, 0);
			if((cur_dev->use_time[j] <= tasks[i].deadline) && 
				(cur_dev_energy <= cur_dev->pow_budget))
			{
				if(cur_dev->use_time[j] < eft)
				{
					eft = cur_dev->use_time[j];
					flag = 0;
					eft_core = j;
					eft_dev = cur_dev;
				}
			}
			cur_dev->use_time[j] -= tasks[i].length / cur_dev->cap;
		}

		
		//scheduling the task to an edge server
		for(j=0; j<num_edge; j++)
		{
			if(cur_dev->edge != j)
				continue;
			for(k=0; k<edges[j].svr_num; k++)
			{
				struct server *cur_srv = edges[j].svr + k;
				for(l=0; l<cur_srv->core; l++)
				{
					cur_dev->updata += tasks[i].sizeIn;
					cur_dev_energy = cal_energy_dev(cur_dev, tasks[i].sizeIn, edges[j].bw_down);
					
					cur_srv->in_time[l] += tasks[i].sizeIn / edges[j].bw_down * 8;
					//the later of the ready time for input data and cpu
					if(cur_srv->in_time[l] > cur_srv->cpu_time[l])
						cur_srv->cpu_time[l] += cur_srv->in_time[l] - cur_srv->cpu_time[l];
					cur_srv->cpu_time[l] += tasks[i].length / cur_srv->cap;
					
					if((cur_srv->cpu_time[l] <= tasks[i].deadline) && 
						(cur_dev_energy <= cur_dev->pow_budget))
					{
						if(cur_srv->cpu_time[l] < eft)
						{
							eft = cur_srv->cpu_time[l];
							flag = 1;
							eft_core = l;
							eft_server = cur_srv;
							eft_edge = edges + j;
						}
					}
					cur_dev->updata -= tasks[i].sizeIn;
					
					cur_srv->in_time[l] -= tasks[i].sizeIn / edges[j].bw_down * 8;
					if(cur_srv->in_time[l] > cur_srv->cpu_time[l])
						cur_srv->cpu_time[l] -= cur_srv->in_time[l] - cur_srv->cpu_time[l];
					cur_srv->cpu_time[l] -= tasks[i].length / cur_srv->cap;
				}
			}

		}


		//scheduling the task to a cloud
		for(j=0; j<num_cloud; j++)
		{
			struct cloud *cur_cloud = clouds + j;
			struct vm *cur_vm = cur_cloud->vms;
			while(cur_vm)
			{
				cur_dev->updata += tasks[i].sizeIn;
				cur_dev_energy = cal_energy_dev(cur_dev, tasks[i].sizeIn, cur_cloud->bw_down);
				
				cur_vm->in_time += tasks[i].sizeIn / cur_cloud->bw_down * 8;
				if(cur_vm->in_time > cur_vm->cpu_time)
					cur_vm->cpu_time += cur_vm->in_time - cur_vm->cpu_time;
				cur_vm->cpu_time += tasks[i].length / cur_cloud->cap;
				
				if((cur_vm->cpu_time <= tasks[i].deadline) && 
					(cur_dev_energy <= cur_dev->pow_budget))
				{
					if(cur_vm->cpu_time < eft)
					{
						eft = cur_vm->cpu_time;
						flag = 2;
						eft_vm = cur_vm;
						eft_cloud = cur_cloud;
					}
				}
				cur_dev->updata -= tasks[i].sizeIn;
				
				cur_vm->cpu_time -= tasks[i].length / cur_cloud->cap;
				if(cur_vm->in_time > cur_vm->cpu_time)
					cur_vm->cpu_time -= cur_vm->in_time - cur_vm->cpu_time;
				cur_vm->in_time -= tasks[i].sizeIn / cur_cloud->bw_down * 8;
				
				cur_vm = cur_vm->next;
			}

		}
		// rent a new VM
		if(flag < 0)
		{
			cur_dev->updata += tasks[i].sizeIn;
			for(j=0; j<num_cloud; j++)
			{
				struct cloud *cur_cloud = clouds + j;
				cur_dev_energy = cal_energy_dev(cur_dev, tasks[i].sizeIn, cur_cloud->bw_down);
				
				float ft = tasks[i].sizeIn / cur_cloud->bw_down * 8 + tasks[i].length / cur_cloud->cap;
				if((ft <= tasks[i].deadline) && (cur_dev_energy <= cur_dev->pow_budget))
				{
					if(ft < eft)
					{
						eft = ft;
						flag = 3;
						eft_cloud = cur_cloud;
					}
				}

			}

			cur_dev->updata -= tasks[i].sizeIn;
		}
		if(flag >= 0)
		{
			rst.fin_num++;
			rst.fin_size += tasks[i].length;
			rst.input += tasks[i].sizeIn;
			switch(flag)
			{
				case 0:
					rst.local_num ++;
					rst.local_size += tasks[i].length;
					rst.local_input += tasks[i].sizeIn;
					eft_dev->use_time[eft_core] += tasks[i].length / eft_dev->cap;
					break;
				case 1:
					rst.edge_num ++;
					rst.edge_size += tasks[i].length;
					rst.edge_input += tasks[i].sizeIn;

					cur_dev->updata += tasks[i].sizeIn;
					cur_dev->updata_edge[eft_edge - edges] += tasks[i].sizeIn;
					
					eft_server->in_time[eft_core] += tasks[i].sizeIn / eft_edge->bw_down * 8;
					eft_server->cpu_time[eft_core] += tasks[i].length / eft_server->cap;
					//the later of the ready time for input data and cpu
					if(eft_server->in_time[eft_core] > eft_server->cpu_time[eft_core])
						eft_server->cpu_time[eft_core] += eft_server->in_time[eft_core] - eft_server->cpu_time[eft_core];
					break;
				case 2:
					rst.cloud_num ++;
					rst.cloud_size += tasks[i].length;
					rst.cloud_input += tasks[i].sizeIn;
					cur_dev->updata_cloud[eft_cloud - clouds] += tasks[i].sizeIn;
					eft_vm->in_time += tasks[i].sizeIn / eft_cloud->bw_down * 8;
					if(eft_vm->in_time > eft_vm->cpu_time)
						eft_vm->cpu_time += eft_vm->in_time - eft_vm->cpu_time;
					eft_vm->cpu_time += tasks[i].length / eft_cloud->cap;
					break;
				case 3:
					rst.cloud_num ++;
					rst.cloud_size += tasks[i].length;
					rst.cloud_input += tasks[i].sizeIn;
					cur_dev->updata_cloud[eft_cloud - clouds] += tasks[i].sizeIn;
					eft_vm = (struct vm *)malloc(sizeof(struct vm));
					memset(eft_vm, 0, sizeof(struct vm));
					eft_vm->cloud = eft_cloud;
					eft_vm->in_time = tasks[i].sizeIn / eft_cloud->bw_down * 8;
					eft_vm->cpu_time = eft_vm->in_time + tasks[i].length / eft_cloud->cap;
					eft_vm->next = eft_cloud->vms;
					eft_cloud->vms = eft_vm;
					break;
				default:
					printf("No such situation: flag = %d\n", flag);
			}
		}
	}
	float max_time = 0.0;
	for(i=0; i<num_dev; i++)
	{
		rst.dev_energy += cal_energy_dev(devs + i, 0.0, 0);
		max_time = find_max_time(devs[i].use_time, devs[i].core);
		rst.local_used += max_time * devs[i].cap * devs[i].core;
		if(rst.ft < max_time)
			rst.ft = max_time;
	}
	for(i=0; i<num_edge; i++)
	{
		for(j=0; j<edges[i].svr_num; j++)
		{
			struct server *cur_srv = edges[i].svr + j;
			max_time = find_max_time(cur_srv->cpu_time, cur_srv->core);
			rst.edge_used += max_time * cur_srv->cap * cur_srv->core;
			if(rst.ft < max_time)
				rst.ft = max_time;
		}
	}
	for(i=0; i<num_cloud; i++)
	{
		rst.cost += cal_cost(clouds + i);
	}
	rst.total_used = rst.local_used + rst.edge_used + rst.cloud_used;
	return rst.fin_num;

}


/***
*** 
*** MTA2019-17-A hybrid approach to scheduling real-time IoT workflows in fog and cloud environments
*** Stavrinides and Karatza \cite{StavrinidesMTA2019} proposed a heuristic method for 
*** improving the SLA violation. Their proposed method respectively employed EDF and 
*** earliest finish time first (EFTF) for task selection and resource allocation, and 
*** tried to fill a task before the input data is ready for the next task to execute.
***/

int edf_eftf()
{
	
	memset(&rst, 0, sizeof(struct result));
	sort_task_deadline_inc();
	return eftf();
}


void end();
