#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <float.h>

#include <time.h>

#include "tasks.h"
#include "resources.h"
#include "result.h"
float cal_ft(struct tsk *task, float cap, float bw, float *ft_cpu, float *ft_input)
{
	if(bw > 1e-4)
		*ft_input += task->sizeIn / bw;
	if(*ft_cpu < *ft_input)
		*ft_cpu = *ft_input;
	*ft_cpu += task->length / cap;
	return *ft_cpu;
}
float cal_slack(struct tsk *task_list, struct tsk *new_task, float cap, float bw)
{
	float cul_slack = 0.0;
	struct tsk *pos;
	float ft_cpu = 0, ft_input = 0;
	if(NULL == task_list)
	{
		if(bw > 1e-4)
			return (new_task->deadline - new_task->length / cap - new_task->sizeIn / bw);
		else
			return new_task->deadline - new_task->length / cap;
	}
	pos = task_list;
	while(pos && new_task->deadline > pos->deadline)
	{
		cal_ft(pos, cap, bw, &ft_cpu, &ft_input);
		if(ft_cpu > pos->deadline)
			return -1.0;
		cul_slack += pos->deadline - ft_cpu;
		pos = pos->next;
	}
	cal_ft(new_task, cap, bw, &ft_cpu, &ft_input);
	if(ft_cpu > new_task->deadline)
		return -1.0;
	cul_slack += new_task->deadline - ft_cpu;
	while(pos)
	{
		cal_ft(pos, cap, bw, &ft_cpu, &ft_input);
		if(ft_cpu > pos->deadline)
			return -1.0;
		cul_slack += pos->deadline - ft_cpu;
		pos = pos->next;
	}
	return cul_slack;
}


void insert_task_dl_inc(struct tsk **task_list, struct tsk *task)
{
	struct tsk *prev, *pos;
	if(NULL == *task_list)
	{
		task->next = NULL;
		*task_list = task;
	}
	else if((*task_list)->deadline > task->deadline)
	{
		task->next = *task_list;
		*task_list = task;
	}
	else
	{
		prev = *task_list;
		pos = prev->next;
		while(pos && pos->deadline < task->deadline)
		{
			prev = pos;
			pos = pos->next;
		}
		task->next = pos;
		prev->next = task;
	}
}

void del_task(struct tsk **task_list, struct tsk *task)
{
	struct tsk *pos = *task_list;
	if(! pos)
		return;
	if(pos == task)
	{
		*task_list = task->next;
		return;
	}
	while(pos && pos->next != task)
		pos = pos->next;
	if(!pos)
	{
		pos->next = task->next;
	}
}

void del_task_dlink(struct tsk **task_list, struct tsk *task)
{
	struct tsk *pos = *task_list;
	if(! pos)
		return;
	if(pos == task && pos->next == *task_list)
	{
		*task_list = NULL;
		return;
	}
	if(pos == task)
	{
		*task_list = task->next;
		pos->next->pre = pos->pre;
		pos->pre->next = pos->next;
		return;
	}
	pos = pos->next;
	while(pos != *task_list && pos != task)
		pos = pos->next;
	if(pos != *task_list)
	{
		pos->next->pre = pos->pre;
		pos->pre->next = pos->next;
	}
}
void insert_task_dl_inc_dlink(struct tsk **task_list, struct tsk *task)
{
	struct tsk *pos;
	if(NULL == *task_list)
	{
		task->next = task;
		task->pre = task;
		*task_list = task;
	}
	else if((*task_list)->deadline > task->deadline)
	{
		task->pre = (*task_list)->pre;
		(*task_list)->pre->next = task;
		task->next = *task_list;
		(*task_list)->pre = task;
		*task_list = task;
	}
	else
	{
		pos = (*task_list)->next;
 		while(pos != *task_list && pos->deadline < task->deadline)
		{
			pos = pos->next;
		}
		task->next = pos;
		task->pre = pos->pre;
		pos->pre->next = task;
		pos->pre = task;
	}
}



void del_vm(struct cloud *cur_cloud, struct vm *cur_vm)
{
	struct vm *pos = cur_cloud->vms;
	if(! pos)
		return;
	if(pos == cur_vm)
	{
		cur_cloud->vms = cur_vm->next;
		return;
	}
	while(pos && pos->next != cur_vm)
		pos = pos->next;
	if(!pos)
	{
		pos->next = cur_vm->next;
	}
	free(cur_vm);
}


void cal_times_vm(struct vm *cvm, float bw)
{
	struct tsk *pos;
	cvm->in_time = 0.0;
	cvm->cpu_time = 0.0;

	pos = cvm->tasks;
	while(pos)
	{
		cvm->in_time += pos->sizeIn / bw * 8;
		if(cvm->in_time > cvm->cpu_time)
			cvm->cpu_time = cvm->in_time + pos->length / cvm->cloud->cap;
		else
			cvm->cpu_time += pos->length / cvm->cloud->cap;
		pos = pos->next;
	}
}

int replacable(struct tsk *task_list, struct tsk *lgest_sch_tsk, 
					struct tsk *smest_unsch_tsk, struct tsk *lgest_unsch_tsk,
					float cap, float bw)
{
	struct tsk *pos;
	float in_time = 0.0;
	float cpu_time = 0.0;
	if(smest_unsch_tsk->deadline > lgest_unsch_tsk->deadline)
	{
		pos = smest_unsch_tsk;
		smest_unsch_tsk = lgest_unsch_tsk;
		lgest_unsch_tsk = pos;
	}

	pos = task_list;
	if(NULL == lgest_sch_tsk)
	{
		while(pos && smest_unsch_tsk->deadline < pos->deadline)
		{
			if(bw > -0.01)
				in_time += pos->sizeIn / bw;
			if(in_time > cpu_time)
				cpu_time = in_time + pos->length / cap;
			else
				cpu_time += pos->length / cap;
			pos = pos->next;
		}
		if(bw > -0.01)
			in_time += smest_unsch_tsk->sizeIn / bw;
		if(in_time > cpu_time)
			cpu_time = in_time + smest_unsch_tsk->length / cap;
		else
			cpu_time += smest_unsch_tsk->length / cap;
		if(cpu_time > smest_unsch_tsk->deadline)
			return 0;
		
		while(pos && lgest_unsch_tsk->deadline < pos->deadline)
		{
			if(bw > -0.01)
				in_time += pos->sizeIn / bw;
			if(in_time > cpu_time)
				cpu_time = in_time + pos->length / cap;
			else
				cpu_time += pos->length / cap;
			if(cpu_time > pos->deadline)
				return 0;
			pos = pos->next;
		}
		if(bw > -0.01)
			in_time += lgest_unsch_tsk->sizeIn / bw;
		if(in_time > cpu_time)
			cpu_time = in_time + lgest_unsch_tsk->length / cap;
		else
			cpu_time += lgest_unsch_tsk->length / cap;
		if(cpu_time > lgest_unsch_tsk->deadline)
			return 0;

		while(pos)
		{
			if(bw > -0.01)
				in_time += pos->sizeIn / bw;
			if(in_time > cpu_time)
				cpu_time = in_time + pos->length / cap;
			else
				cpu_time += pos->length / cap;
			if(cpu_time > pos->deadline)
				return 0;
			pos = pos->next;
		}
		return 1;
		
	}
	
	while(pos && pos != lgest_sch_tsk && smest_unsch_tsk->deadline < pos->deadline)
	{
		if(bw > -0.01)
			in_time += pos->sizeIn / bw;
		if(in_time > cpu_time)
			cpu_time = in_time + pos->length / cap;
		else
			cpu_time += pos->length / cap;
		pos = pos->next;
	}
	if(!pos)
	{
		printf("No possible as there must be lgest_sch_tsk in the list\n");
		return 0;
	}
 	else if(pos == lgest_sch_tsk)
 	{
		pos = pos->next;
		while(pos && smest_unsch_tsk->deadline < pos->deadline)
		{
			if(bw > -0.01)
				in_time += pos->sizeIn / bw;
			if(in_time > cpu_time)
				cpu_time = in_time + pos->length / cap;
			else
				cpu_time += pos->length / cap;
			pos = pos->next;
		}
		if(bw > -0.01)
			in_time += smest_unsch_tsk->sizeIn / bw;
		if(in_time > cpu_time)
			cpu_time = in_time + smest_unsch_tsk->length / cap;
		else
			cpu_time += smest_unsch_tsk->length / cap;
		if(cpu_time > smest_unsch_tsk->deadline)
			return 0;
		
		while(pos && lgest_unsch_tsk->deadline < pos->deadline)
		{
			if(bw > -0.01)
				in_time += pos->sizeIn / bw;
			if(in_time > cpu_time)
				cpu_time = in_time + pos->length / cap;
			else
				cpu_time += pos->length / cap;
			if(cpu_time > pos->deadline)
				return 0;
			pos = pos->next;
		}
		if(bw > -0.01)
			in_time += lgest_unsch_tsk->sizeIn / bw;
		if(in_time > cpu_time)
			cpu_time = in_time + lgest_unsch_tsk->length / cap;
		else
			cpu_time += lgest_unsch_tsk->length / cap;
		if(cpu_time > lgest_unsch_tsk->deadline)
			return 0;

		while(pos)
		{
			if(bw > -0.01)
				in_time += pos->sizeIn / bw;
			if(in_time > cpu_time)
				cpu_time = in_time + pos->length / cap;
			else
				cpu_time += pos->length / cap;
			if(cpu_time > pos->deadline)
				return 0;
			pos = pos->next;
		}
		
 	}
	else //smest_unsch_tsk->deadline < pos->deadline
	{
		if(bw > -0.01)
			in_time += smest_unsch_tsk->sizeIn / bw;
		if(in_time > cpu_time)
			cpu_time = in_time + smest_unsch_tsk->length / cap;
		else
			cpu_time += smest_unsch_tsk->length / cap;
		if(cpu_time > smest_unsch_tsk->deadline)
			return 0;
		while(pos && pos != lgest_sch_tsk && lgest_unsch_tsk->deadline < pos->deadline)
		{
			if(bw > -0.01)
				in_time += pos->sizeIn / bw;
			if(in_time > cpu_time)
				cpu_time = in_time + pos->length / cap;
			else
				cpu_time += pos->length / cap;
			if(cpu_time > pos->deadline)
				return 0;
			pos = pos->next;
		}
		if(!pos)
		{
			if(bw > -0.01)
				in_time += lgest_unsch_tsk->sizeIn / bw;
			if(in_time > cpu_time)
				cpu_time = in_time + lgest_unsch_tsk->length / cap;
			else
				cpu_time += lgest_unsch_tsk->length / cap;
			if(cpu_time > lgest_unsch_tsk->deadline)
				return 0;
		}
		else if(pos == lgest_sch_tsk)

		{
			pos = pos->next;
			while(pos && lgest_unsch_tsk->deadline < pos->deadline)
			{
				if(bw > -0.01)
					in_time += pos->sizeIn / bw;
				if(in_time > cpu_time)
					cpu_time = in_time + pos->length / cap;
				else
					cpu_time += pos->length / cap;
				if(cpu_time > pos->deadline)
					return 0;
				pos = pos->next;
			}
			if(bw > -0.01)
				in_time += lgest_unsch_tsk->sizeIn / bw;
			if(in_time > cpu_time)
				cpu_time = in_time + lgest_unsch_tsk->length / cap;
			else
				cpu_time += lgest_unsch_tsk->length / cap;
			if(cpu_time > lgest_unsch_tsk->deadline)
				return 0;

			while(pos)
			{
				if(bw > -0.01)
					in_time += pos->sizeIn / bw;
				if(in_time > cpu_time)
					cpu_time = in_time + pos->length / cap;
				else
					cpu_time += pos->length / cap;
				if(cpu_time > pos->deadline)
					return 0;
				pos = pos->next;
			}
			
		}
		else //lgest_unsch_tsk->deadline > pos->deadline
		{
			if(bw > -0.01)
				in_time += lgest_unsch_tsk->sizeIn / bw;
			if(in_time > cpu_time)
				cpu_time = in_time + lgest_unsch_tsk->length / cap;
			else
				cpu_time += lgest_unsch_tsk->length / cap;
			if(cpu_time > lgest_unsch_tsk->deadline)
				return 0;

			while(pos)
			{
				if(pos == lgest_sch_tsk)
				{
					pos = pos->next;
					continue;
				}
				if(bw > -0.01)
					in_time += pos->sizeIn / bw;
				if(in_time > cpu_time)
					cpu_time = in_time + pos->length / cap;
				else
					cpu_time += pos->length / cap;
				if(cpu_time > pos->deadline)
					return 0;
				pos = pos->next;
			}
		}
	}
	return 1;
}


void cal_times(struct server *svr, float bw)
{
	struct tsk *pos;
	int i;
	memset(svr->in_time, 0, MAX_CORE_PER_NODE * sizeof(float));
	memset(svr->cpu_time, 0, MAX_CORE_PER_NODE * sizeof(float));
	for(i=0; i<svr->core; i++)
	{
		pos = svr->tasks[i];
		while(pos)
		{
			svr->in_time[i] += pos->sizeIn / bw * 8;
			if(svr->in_time[i] > svr->cpu_time[i])
				svr->cpu_time[i] = svr->in_time[i] + pos->length / svr->cap;
			else
				svr->cpu_time[i] += + pos->length / svr->cap;
			pos = pos->next;
		}
	}
}

int heuristic()
{
	
	memset(&rst, 0, sizeof(struct result));
	time_t start_time = time(NULL);
	int i_task, i_cloud, i_edge, i_svr, i_core;
	struct tsk *cur_task;
	struct dev *cur_dev;
	struct edge *cur_edge;
	struct server *cur_svr;
	struct cloud *cur_cloud;
	float cur_dev_energy;
	char sch_flags[MAX_TASKS];
	memset(sch_flags, 0, MAX_TASKS * sizeof(char));

	//first round for trying to complete all tasks
//	printf("first round for trying to complete all tasks\n");
	for(i_task=0; i_task<num_task; i_task++)
	{
		cur_task = tasks + i_task;
		cur_dev = cur_task->dev;
		/**
		**  cloud resources
		**/
		for(i_cloud=0; i_cloud<num_cloud; i_cloud++)
		{
			cur_cloud = clouds + i_cloud;
			cur_dev->updata += cur_task->sizeIn;
			cur_dev_energy = cal_energy_dev(cur_dev, cur_task->sizeIn, cur_cloud->bw_down);
			
			float ft = cur_task->sizeIn / cur_cloud->bw_down * 8 + cur_task->length / cur_cloud->cap;
			if((ft <= cur_task->deadline) && (cur_dev_energy <= cur_dev->pow_budget))
			{
				rst.cloud_num ++;
				rst.cloud_size += cur_task->length;
				rst.fin_num ++;
				rst.fin_size += cur_task->length;
				rst.cloud_input += cur_task->sizeIn;
				rst.input += cur_task->sizeIn;
				cur_dev->updata_cloud[i_cloud] += cur_task->sizeIn;
				struct vm *cur_vm = (struct vm *)malloc(sizeof(struct vm));
				if(! cur_vm)
				{
					printf("ERROR malloc for a new vm\n");
					exit(0);

				}
				memset(cur_vm, 0, sizeof(struct vm));
				cur_vm->in_time = cur_task->sizeIn / cur_cloud->bw_down * 8;
				cur_vm->cpu_time = cur_vm->in_time + cur_task->length / cur_cloud->cap;
				cur_vm->cloud = cur_cloud;
				insert_task_dl_inc(&cur_vm->tasks, cur_task);
				//insert vm in the vm list
				struct vm *tmp;
				if(! cur_cloud->vms)
					 cur_cloud->vms = cur_vm;
				else
				{
					tmp = cur_cloud->vms;
					while(tmp->next)
						tmp = tmp->next;
					tmp->next = cur_vm;
					cur_vm->next = NULL;
				}
				sch_flags[cur_task - tasks] = 1;
				goto next_task;
			}

			else
			{
				cur_dev->updata -= cur_task->sizeIn;
			}
		}
		//Now, no task can be finished in the cloud.
		//edge
		float min_slack = FLT_MAX;
		int min_core = -1;
		struct edge *min_edge = NULL;
		struct server *min_svr = NULL;
		for(i_edge=0; i_edge<num_edge; i_edge++)
		{
			if(cur_dev->edge != i_edge)
				continue;
			cur_edge = edges + i_edge;
			for(i_svr=0; i_svr<cur_edge->svr_num; i_svr++)
			{
				cur_svr = cur_edge->svr + i_svr;
				for(i_core=0; i_core < cur_svr->core; i_core++)
				{
					float acc_slack;
					acc_slack = cal_slack(cur_svr->tasks[i_core], cur_task, cur_svr->cap, cur_edge->bw_down);
					if(acc_slack > -0.1 && min_slack > acc_slack)
					{
						min_slack = acc_slack;
						min_edge = cur_edge;
						min_svr = cur_svr;
						min_core = i_core;
					}
				}
			}
		}
		if(min_edge)
		{
			insert_task_dl_inc(&min_svr->tasks[min_core], cur_task);
			cur_dev->updata += cur_task->sizeIn;
			min_svr->in_time[min_core] += cur_task->sizeIn / min_edge->bw_down * 8;
			if(min_svr->in_time[min_core] > min_svr->cpu_time[min_core])
				min_svr->cpu_time[min_core] = min_svr->in_time[min_core];
			min_svr->cpu_time[min_core] += cur_task->length / min_svr->cap;
			
			rst.edge_num ++;
			rst.edge_size += cur_task->length;
			rst.fin_num ++;
			rst.fin_size += cur_task->length;
			rst.edge_input += cur_task->sizeIn;
			rst.input += cur_task->sizeIn;
			cur_dev->updata_edge[min_edge - edges] += cur_task->sizeIn;
			sch_flags[cur_task - tasks] = 1;
			goto next_task;
		}
		//devices
		min_core = -1;
		min_slack = FLT_MAX;
		for(i_core=0; i_core < cur_dev->core; i_core++)
		{
			float acc_slack;
			acc_slack = cal_slack(cur_dev->tasks[i_core], cur_task, cur_dev->cap, -1);
			if(acc_slack > -0.1 && min_slack > acc_slack)
			{
				min_slack = acc_slack;
				min_core = i_core;
			}
		}
		if(min_core >= 0)
		{
			insert_task_dl_inc(&cur_dev->tasks[min_core], cur_task);
			cur_dev->use_time[min_core] += cur_task->length / cur_dev->cap;
			rst.local_num ++;
			rst.local_size += cur_task->length;
			rst.fin_num ++;
			rst.fin_size += cur_task->length;
			rst.local_input += cur_task->sizeIn;
			rst.input += cur_task->sizeIn;
			sch_flags[cur_task - tasks] = 1;
		}
		next_task:
		;
	}
	char flag = 0;
	struct tsk *pos, *pre;
	do{
		flag = 0;
		//reschedule tasks from edge to device, to make room for remaning resources
//	printf("reschedule tasks from edge to device, to make room for remaning resources\n");
		for(i_edge=0; i_edge<num_edge; i_edge++)
		{
			cur_edge = edges + i_edge;
			for(i_svr=0; i_svr<cur_edge->svr_num; i_svr++)
			{
				cur_svr = cur_edge->svr + i_svr;
				for(i_core=0; i_core < cur_svr->core; i_core++)
				{
					pos = cur_svr->tasks[i_core];
					while(pos)
					{
						int min_core = -1;
						float min_slack = FLT_MAX;
						int i_core_dev;
						float acc_slack;
						cur_dev = pos->dev;
						for(i_core_dev=0; i_core_dev < cur_dev->core; i_core_dev++)
						{
							acc_slack = cal_slack(cur_dev->tasks[i_core_dev], pos, cur_dev->cap, -1);
							if(acc_slack > -0.1 && min_slack > acc_slack)
							{
								min_slack = acc_slack;
								min_core = i_core_dev;
							}
						}
						if(min_core >= 0)
						{
							insert_task_dl_inc(&cur_dev->tasks[min_core], pos);
							cur_dev->use_time[min_core] += pos->length / cur_dev->cap;
							rst.local_num ++;
							rst.local_size += pos->length;
							rst.local_input += pos->sizeIn;
							rst.edge_num --;
							rst.edge_size -= pos->length;
							rst.edge_input -= pos->sizeIn;
							cur_dev->updata_edge[cur_edge - edges] -= pos->sizeIn;
							pre = pos;
							pos = pos->next;
							del_task(& cur_svr->tasks[i_core], pre);
							cal_times(cur_svr, cur_edge->bw_down);
						}
						else
							pos = pos->next;
					}
				}
			}
		}
		//schedule tasks to edge, after this, there is no task can be finished any more
//		printf("schedule tasks to edge, after this, there is no task can be finished any more\n");
		for(i_task=0; i_task<num_task && ! sch_flags[i_task]; i_task++)
		{
			cur_task = tasks + i_task;
			cur_dev = cur_task->dev;
		
			float min_slack = FLT_MAX;
			int min_core = -1;
			struct edge *min_edge = NULL;
			struct server *min_svr = NULL;
			for(i_edge=0; i_edge<num_edge; i_edge++)
			{
				if(cur_dev->edge != i_edge)
					continue;
				cur_edge = edges + i_edge;
				for(i_svr=0; i_svr<cur_edge->svr_num; i_svr++)
				{
					cur_svr = cur_edge->svr + i_svr;
					for(i_core=0; i_core < cur_svr->core; i_core++)
					{
						float acc_slack;
						acc_slack = cal_slack(cur_svr->tasks[i_core], cur_task, cur_svr->cap, cur_edge->bw_down);
						if(acc_slack > -0.1 && min_slack > acc_slack)
						{
							min_slack = acc_slack;
							min_edge = cur_edge;
							min_svr = cur_svr;
							min_core = i_core;
						}
					}
				}
			}
			if(min_edge)
			{
				insert_task_dl_inc(&min_svr->tasks[min_core], cur_task);
				cur_dev->updata += cur_task->sizeIn;
				min_svr->in_time[min_core] += cur_task->sizeIn / min_edge->bw_down * 8;
				if(min_svr->in_time[min_core] > min_svr->cpu_time[min_core])
					min_svr->cpu_time[min_core] = min_svr->in_time[min_core];
				min_svr->cpu_time[min_core] += cur_task->length / min_svr->cap;
				
				rst.edge_num ++;
				rst.edge_size += cur_task->length;
				rst.fin_num ++;
				rst.fin_size += cur_task->length;
				rst.edge_input += cur_task->sizeIn;
				rst.input += cur_task->sizeIn;
				cur_dev->updata_edge[min_edge - edges] += cur_task->sizeIn;
				sch_flags[cur_task - tasks] = 1;
			}
		}


//		printf("remain unscheduled tasks\n");
		if(0)//if(rst.fin_num < num_task)
		{//remain unscheduled tasks
			struct tsk unsch_tasks[MAX_TASKS];
			memset(unsch_tasks, 0, MAX_TASKS * sizeof(struct tsk));
			int unsch_num = 0;
			for(i_task=0; i_task<num_task; i_task++)
			{
				if(!sch_flags[i_task])
				{
					memcpy(unsch_tasks + unsch_num, tasks + i_task, sizeof(struct tsk));
					unsch_num ++;
				}
			}
//			printf("%d unscheduled tasks\n", unsch_num);
			sort_task_length_dec_part(unsch_tasks, unsch_num);
			//Bidirectional circular linked list for unscheduled tasks
			for(i_task=1; i_task<unsch_num-1; i_task++)
			{
				unsch_tasks[i_task].next = unsch_tasks + i_task + 1;
				unsch_tasks[i_task].pre = unsch_tasks + i_task - 1;
			}
			unsch_tasks[0].pre = unsch_tasks + i_task;
			unsch_tasks[0].next = unsch_tasks + 1;
			unsch_tasks[i_task].pre = unsch_tasks + i_task - 1;
			unsch_tasks[i_task].next = unsch_tasks;
			//replace one large with two small tasks. In this step, the tasks assigned to the cloud are not searched as
			// no remain task can be finished in the cloud after the first step.
			// largest tasks --> smallest task first + largest task first.
			// repeat until there is no replacement.
			struct tsk *lgest_sch_tsk = NULL;
			struct tsk *lgest_unsch_tsk = NULL;
			struct tsk *smest_unsch_tsk = NULL;
			struct tsk *unsch_list = unsch_tasks;
			int first_flag = 1;
			while(unsch_list != unsch_tasks || (first_flag && unsch_list == unsch_tasks))
			{
				first_flag = 0;
				smest_unsch_tsk = unsch_list;
				cur_dev = smest_unsch_tsk->dev;
				for(i_core = 0; i_core < cur_dev->core; i_core ++)
				{
					// find the scheduled task with largest length
					lgest_sch_tsk = cur_dev->tasks[i_core];
					if(lgest_sch_tsk && lgest_sch_tsk->next)
					{
						pos = cur_dev->tasks[i_core]->next;
						while(pos)
						{
							if(lgest_sch_tsk->length < pos->length)
								lgest_sch_tsk = pos;
							pos = pos->next;
						}
					}
					//if the smallest unsched task with another unsched task cannot replace the largest sched task,
					//more tasks can be finished by the replace operator in this core.
					for(lgest_unsch_tsk = unsch_list->pre;
						lgest_unsch_tsk != smest_unsch_tsk;
						lgest_unsch_tsk = lgest_unsch_tsk->pre)
					{
						if(replacable(cur_dev->tasks[i_core], lgest_sch_tsk, 
							smest_unsch_tsk, lgest_unsch_tsk,
							cur_dev->cap, -1))
						{
							del_task(& cur_dev->tasks[i_core], lgest_sch_tsk);
							del_task_dlink(&unsch_list, smest_unsch_tsk);
							del_task_dlink(&unsch_list, lgest_unsch_tsk);
							insert_task_dl_inc_dlink(&unsch_list, lgest_sch_tsk);
							insert_task_dl_inc(& cur_dev->tasks[i_core], smest_unsch_tsk);
							insert_task_dl_inc(& cur_dev->tasks[i_core], lgest_unsch_tsk);
							float add_length = lgest_unsch_tsk->length + smest_unsch_tsk->length - lgest_sch_tsk->length;
							float add_sizeIn = lgest_unsch_tsk->sizeIn + smest_unsch_tsk->sizeIn - lgest_sch_tsk->sizeIn;
							cur_dev->use_time[i_core] += add_length / cur_dev->cap;
							rst.local_num ++;
							rst.local_size += add_length;
							rst.fin_num ++;
							rst.fin_size += add_length;
							rst.local_input += add_sizeIn;
							rst.input += add_sizeIn;
							printf("%d\n", rst.fin_num);
							goto next_unsch_task;
						}
					}
				}
				next_unsch_task:
				unsch_list = unsch_list->next;
			}
			//replace smallest with largest task, for improving finished length/size.
		}
		
//		printf("reschedule tasks from clouds to device, as the network trans power is high.\n");
		//reschedule tasks from clouds to device, as the network trans power is high.
		// After that, more tasks may be finished due to more remain energy budget
		for(i_cloud=0; i_cloud<num_cloud; i_cloud++)
		{
			cur_cloud = clouds + i_cloud;
			struct vm *cur_vm = cur_cloud->vms;
			while(cur_vm)
			{
				cur_task = cur_vm->tasks;
				while(cur_task)
				{
					int min_core = -1;
					float min_slack = FLT_MAX;
					int i_core_dev;
					float acc_slack;
					cur_dev = cur_task->dev;
					for(i_core_dev=0; i_core_dev < cur_dev->core; i_core_dev++)
					{
						acc_slack = cal_slack(cur_dev->tasks[i_core_dev], cur_task, cur_dev->cap, -1);
						if(acc_slack > -0.1 && min_slack > acc_slack)
						{
							min_slack = acc_slack;
							min_core = i_core_dev;
						}
					}
					if(min_core >= 0)
					{
						pre = cur_task;
						cur_task = cur_task->next;
						del_task(& cur_vm->tasks, pre);
						insert_task_dl_inc(&cur_dev->tasks[min_core], pre);
						cur_dev->use_time[min_core] += pre->length / cur_dev->cap;
						rst.local_num ++;
						rst.local_size += pre->length;
						rst.local_input += pre->sizeIn;
						rst.cloud_num --;
						rst.cloud_size -= pre->length;
						rst.cloud_input -= pre->sizeIn;
						cur_dev->updata_cloud[cur_cloud - clouds] -= pre->sizeIn;
						cal_times_vm(cur_vm, cur_cloud->bw_down);
					}
					else
						cur_task = cur_task->next;
				}
				cur_vm = cur_vm->next;
			}
		}

		//reschedule tasks from clouds to edge 
//		printf("reschedule tasks from clouds to edge .\n");
		for(i_cloud=0; i_cloud<num_cloud; i_cloud++)
		{
			cur_cloud = clouds + i_cloud;
			struct vm *cur_vm = cur_cloud->vms;
			while(cur_vm)
			{
				cur_task = cur_vm->tasks;
				while(cur_task)
				{
					cur_dev = cur_task->dev;
		
					float min_slack = FLT_MAX;
					int min_core = -1;
					struct edge *min_edge = NULL;
					struct server *min_svr = NULL;
					for(i_edge=0; i_edge<num_edge; i_edge++)
					{
						if(cur_dev->edge != i_edge)
							continue;
						cur_edge = edges + i_edge;
						for(i_svr=0; i_svr<cur_edge->svr_num; i_svr++)
						{
							cur_svr = cur_edge->svr + i_svr;
							for(i_core=0; i_core < cur_svr->core; i_core++)
							{
								float acc_slack;
								acc_slack = cal_slack(cur_svr->tasks[i_core], cur_task, cur_svr->cap, cur_edge->bw_down);
								if(acc_slack > -0.1 && min_slack > acc_slack)
								{
									min_slack = acc_slack;
									min_edge = cur_edge;
									min_svr = cur_svr;
									min_core = i_core;
								}
							}
						}
					}
					if(min_edge)
					{
						pre = cur_task;
						cur_task = cur_task->next;
						del_task(& cur_vm->tasks, pre);
						cal_times_vm(cur_vm, cur_cloud->bw_down);
						insert_task_dl_inc(&min_svr->tasks[min_core], pre);
						cur_dev->updata += pre->sizeIn;
						min_svr->in_time[min_core] += pre->sizeIn / min_edge->bw_down * 8;
						if(min_svr->in_time[min_core] > min_svr->cpu_time[min_core])
							min_svr->cpu_time[min_core] = min_svr->in_time[min_core];
						min_svr->cpu_time[min_core] += pre->length / min_svr->cap;
						
						rst.edge_num ++;
						rst.edge_size += pre->length;
						rst.edge_input += pre->sizeIn;
						cur_dev->updata_edge[min_edge - edges] += pre->sizeIn;
						rst.cloud_num --;
						rst.cloud_size -= pre->length;
						rst.cloud_input -= pre->sizeIn;
						cur_dev->updata_cloud[cur_cloud - clouds] -= pre->sizeIn;
					}
					else
						cur_task = cur_task->next;
				}
				cur_vm = cur_vm->next;
			}
		}
		flag = 0;
	}while(flag);

//	printf("rescheduling tasks in clouds to multiplex VMs.\n");
	
	//rescheduling tasks in clouds to multiplex VMs
	for(i_cloud = 0; i_cloud < num_cloud; i_cloud++)
	{
		cur_cloud = clouds + i_cloud;
		struct vm *cur_vm = cur_cloud->vms;
		struct vm *pre_vm = NULL, *pos_vm = NULL;
		pos_vm = cur_vm = cur_cloud->vms;
		while(cur_vm && pos_vm)
		{
			pos_vm = cur_vm->next;
			while(pos_vm)
			{
				if(! pos_vm->tasks)
				{
					pre_vm = pos_vm;
					pos_vm = pos_vm->next;
					del_vm(cur_cloud, pre_vm);
					continue;
				}
				if(cal_slack(cur_vm->tasks, pos_vm->tasks, cur_cloud->cap, cur_cloud->bw_down) > -0.1)
				{
					insert_task_dl_inc(&cur_vm->tasks, pos_vm->tasks);
					cal_times_vm(cur_vm, cur_cloud->bw_down);
				}
				pre_vm = pos_vm;
				pos_vm = pos_vm->next;
				del_vm(cur_cloud, pre_vm);
			}
			cur_vm = cur_vm->next;
		}
	}
	start_time -= time(NULL);
	
//	printf("scheduling over.\n");
	float max_time = 0.0;
	int i_dev;
	for(i_dev=0; i_dev<num_dev; i_dev++)
	{
		rst.dev_energy += cal_energy_dev(devs + i_dev, 0.0, 0);
		if(rst.dev_energy < 0)
		{
			printf("XXX===%f\n", devs[i_dev].updata);
			printf("TTT===%f\t%f\t%f\t%f\n", devs[i_dev].use_time[0], devs[i_dev].use_time[1],
				devs[i_dev].use_time[2], devs[i_dev].use_time[3]);
		}
		max_time = find_max_time(devs[i_dev].use_time, devs[i_dev].core);
		rst.local_used += max_time * devs[i_dev].cap * devs[i_dev].core;
		if(rst.ft < max_time)
			rst.ft = max_time;
	}
	for(i_edge=0; i_edge<num_edge; i_edge++)
	{
		for(i_svr=0; i_svr<edges[i_edge].svr_num; i_svr++)
		{
			cur_svr = edges[i_edge].svr + i_svr;
			max_time = find_max_time(cur_svr->cpu_time, cur_svr->core);
			rst.edge_used += max_time * cur_svr->cap * cur_svr->core;
			if(rst.ft < max_time)
				rst.ft = max_time;
		}
	}
	for(i_cloud=0; i_cloud<num_cloud; i_cloud++)
	{
		rst.cost += cal_cost(clouds + i_cloud);
	}
	rst.total_used = rst.local_used + rst.edge_used + rst.cloud_used;
	return rst.fin_num;
}
void end();
