#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "resources.h"
#include "result.h"


int num_dev = 10;
struct dev devs[MAX_DEV];
int dev_energy_budget = 5000;


int num_edge = 1;
struct edge edges[MAX_EDGE];
int radius = 500;//m

int num_cloud = 1;
struct cloud clouds[MAX_CLOUD];


void init_dev()
{
	memset(devs, 0, MAX_DEV * sizeof(struct dev));
	int i;
	for(i=0; i<num_dev; i++)
	{
		devs[i].cap = 1.0 * rand() / RAND_MAX + 1; // [1, 2] GHz
		devs[i].core = rand() % 4 + 1; //1, 2, 3, 4
		devs[i].edge = RAND_MAX % num_edge;
		devs[i].lat = rand() % radius + 1;
		devs[i].lon = rand() % radius + 1;
		devs[i].mem = 1536.0 * rand() / RAND_MAX + 512; // [0.5, 2] GB
		//CJ-CN2020-12-Near-optimal and learning-driven task offloading in a 5G multi-cell mobile edge cloud
		devs[i].pow_busy = 0.5;
		devs[i].pow_idle = 0.05;
		devs[i].pow_trans = 0.2;
		//An Effective Offloading Trade-Off to Economize Energy and Performance in Edge Computing
		devs[i].pow_budget = 5000000; //J, 50\% full charge
	}
}

int find_maxcap_server(struct server *svrs, int srv_num)
{
	int i, i_max;
	float max;
	i_max = 0;
	max = svrs[0].cap * svrs[0].core;
	for(i=1; i<srv_num; i++)
	{
		if(max < svrs[i].cap * svrs[i].core)
		{
			max = svrs[i].cap * svrs[i].core;
			i_max = i;
		}
	}
	return i_max;
}

void init_edge()
{
	memset(edges, 0, MAX_EDGE * sizeof(struct edge));
	int i, j;
	for(i=0; i<num_dev; i++)
	{
		edges[i].svr_num = rand() % 4 + 1; //1, 2, 3, 4
		for(j=0; j<edges[i].svr_num; j++)
		{
			edges[i].svr[j].cap = 1.0 * rand() / RAND_MAX + 2; // [2, 3] GHz;
			edges[i].svr[j].core = rand() % 5 + 4; //4, 5, 6, 7, 8
			edges[i].svr[j].mem = (12.0 * rand() / RAND_MAX + 4) * 1024; // [4, 16] GB
			if(edges[i].svr[j].core <= 6)
			{
				edges[i].svr[j].pow_idle = 110;
				edges[i].svr[j].pow_busy = 175;
				edges[i].svr[j].pow_trans = 100;
			}
			else if(edges[i].svr[j].core <= 12)
			{
				edges[i].svr[j].pow_idle = 125;
				edges[i].svr[j].pow_busy = 210;
				edges[i].svr[j].pow_trans = 100;
			}
			else
			{
				edges[i].svr[j].pow_idle = 210;
				edges[i].svr[j].pow_busy = 300;
				edges[i].svr[j].pow_trans = 100;
			}
		}
//		edges[i].bw_up = 90.0 * rand() / RAND_MAX + 10; //10~100 Mbps
		edges[i].bw_down = 90.0 * rand() / RAND_MAX + 10; //10~100 Mbps
		edges[i].lat = rand() % radius + 1;
		edges[i].lon = rand() % radius + 1;
	}
}

void init_cloud()
{
	memset(clouds, 0, MAX_CLOUD * sizeof(struct cloud));
	int i;
	for(i=0; i<num_cloud; i++)
	{
		clouds[i].cap = 1.0 * rand() / RAND_MAX + 2; // [2, 3] GHz;
		clouds[i].price = 0.01; //$/hour
//		clouds[i].bw_up = 9.0 * rand() / RAND_MAX + 1; //1~10 Mbps
		clouds[i].bw_down = 9.0 * rand() / RAND_MAX + 1; //1~10 Mbps
	}
}

float cal_cost(struct cloud *cloud)
{
	float cost = 0.0;
	struct vm *vm;
	int use_time = 0;
	for(vm=cloud->vms; vm; vm=vm->next)
	{
		use_time = ceil(vm->cpu_time / 3600);
		cost += use_time * cloud->price;
		rst.cloud_used += use_time * 3600 * cloud->cap;
	}
	return cost;
}

void init_resources()
{
	srand( time(NULL) );
	init_dev();
	init_edge();
	init_cloud();
}

int compare_float_inc(const void *p1, const void *p2)
{
	float *tsk1, *tsk2;
	tsk1 = (float *)p1;
	tsk2 = (float *)p2;
	if(*tsk1 - *tsk2 > 0)
		return 1;
	if(*tsk1 - *tsk2 < 0)
		return -1;
	return 0;
}
void sort_float_inc(float *nums, int num)
{
	qsort(nums, num, sizeof(float), compare_float_inc);
}


float cal_energy_dev(struct dev *device, float data, float bw)
{
	int i, j;
	float use_times[MAX_CORE_PER_NODE], energy;
	energy = 0.0;
	memcpy(use_times, device->use_time, MAX_CORE_PER_NODE * sizeof(float));
	sort_float_inc(use_times, device->core);
	for(i = 0; i < device->core; i++)
	{
		if(0 == i)
			energy = use_times[i] * device->pow_busy;
		else
			energy += (use_times[i] - use_times[i-1]) 
				* ((device->pow_busy - device->pow_idle) * (device->core - i) / device->core 
					+ device->pow_idle);
	}
	for(j=0; j<num_edge; j++)
	{
		energy += device->updata_edge[j] / edges[j].bw_down * 8 * device->pow_trans;
	}
	for(j=0; j<num_cloud; j++)
	{
		energy += device->updata_cloud[j] / clouds[j].bw_down * 8 * device->pow_trans;
	}
	if(bw > 1e-4)
		energy += data / bw * 8 * device->pow_trans;
	return energy;
}

float find_max_time(float *times, int n)
{
	int i;
	float max;
	max = times[0];
	for(i=1; i<n; i++)
	{
		if(max < times[i])
		{
			max = times[i];
		}
	}
	return max;
}

