#ifndef _RESOURCES_H_
#define _RESOURCES_H_
#include "tasks.h"

#define MAX_CORE_PER_NODE 32

#define MAX_DEV 1000
#define MAX_EDGE 10
#define MAX_SERVER_ONE_EDGE 5
#define MAX_CLOUD 10
// user device
struct dev{
	float cap; //computing capacity of each core, Hz
	int core; //core number
	float mem; //MB
	float lon;//Longitude
	float lat;//latitude
	int edge; //which edge covering the device
	float pow_idle; //W
	float pow_busy; //W
	float pow_trans; //W
	float pow_budget; // J = W * s
	struct tsk *tasks[MAX_CORE_PER_NODE];
	float use_time[MAX_CORE_PER_NODE];
	//for data transfer energy calculation
	float updata;
	float updata_edge[MAX_EDGE];
	float updata_cloud[MAX_CLOUD];
//	float downdata;
	//
	float energy;
};
extern struct dev devs[MAX_DEV];
extern int num_dev;
extern int dev_energy_budget; //W

// edge server
struct server{
	float cap; //computing capacity of each core, Hz
	int core; //core number
	float mem; //MB
	float pow_idle; //W
	float pow_busy; //W
	float pow_trans; //W
	struct tsk *tasks[MAX_CORE_PER_NODE];
//	float use_time[MAX_CORE_PER_NODE];
	float in_time[MAX_CORE_PER_NODE];
	float cpu_time[MAX_CORE_PER_NODE];
//	float fin_length[MAX_CORE_PER_NODE];
//	float out_time[MAX_CORE_PER_NODE];
};

struct edge{
	struct server svr[MAX_SERVER_ONE_EDGE]; //servers
	int svr_num; //the number of servers in the edge
//	float bw_up; //bandwidth
	float bw_down; //bandwidth
	float lon;//Longitude
	float lat;//latitude
};
//H: 500m; L: 30m radius
//in 2017-11-A strategy for joint service offloading and scheduling in heterogeneous cloud radio access networks
//(EURASIP Journal on Wireless Communications and Networking)
extern int radius; // the cover circle area of an edge in meter

extern struct edge edges[MAX_EDGE];
extern int num_edge;


struct vm{
	struct tsk *tasks;
	float in_time;
	float cpu_time;
	struct cloud *cloud;
	struct vm *next;
};


struct cloud{//vm types
	float cap; //computing capacity of a core, Hz
	float price; //price of a core in each hour
	float bw_down; //bandwidth
	struct vm *vms;
	struct tsk *tasks;
};

extern struct cloud clouds[MAX_CLOUD];
extern int num_cloud;


void init_dev();
void init_edge();
void init_cloud();
void init_resources();
int find_maxcap_server(struct server *svrs, int srv_num);

float cal_energy_dev(struct dev *device, float data, float bw);
float cal_cost(struct cloud *cloud);

float find_max_time(float *times, int n);

#endif
