#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <float.h>
#include <stdio.h>
#include "tasks.h"
#include "resources.h"
#include "result.h"
#include "smallest_first.h"

int smallest_first(enum small_metric metric)
{
	memset(&rst, 0, sizeof(struct result));
	int i, j, k, l;
	double cur_dev_energy;
	float smallest = FLT_MAX;
	float mv = FLT_MAX;
	float add_energy = FLT_MAX;
	int flag;//0 for dev, 1 for edge, 2 for cloud
	int eft_core = -1; //for dev or server
	struct dev *eft_dev = NULL;
	struct edge *eft_edge = NULL;//for edge
	struct server *eft_server = NULL;//for edge
	struct vm *eft_vm = NULL;//for cloud
	struct cloud *eft_cloud = NULL; // for new vm
	for(i=0; i<num_task; i++)
	{
		smallest = FLT_MAX;
		flag = -1;
		struct dev *cur_dev = tasks[i].dev;
		//scheduling the task to the device
		for(j=0; j<cur_dev->core; j++)
		{
			add_energy = cal_energy_dev(cur_dev, 0.0, 0);
			cur_dev->use_time[j] += tasks[i].length / cur_dev->cap;
			cur_dev_energy = cal_energy_dev(cur_dev, 0.0, 0);
			add_energy = cur_dev_energy - add_energy;
			if((cur_dev->use_time[j] <= tasks[i].deadline) && 
				(cur_dev_energy <= cur_dev->pow_budget))
			{
				switch (metric)
				{
					case slack:
						mv = tasks[i].deadline - cur_dev->use_time[j];
						break;
					case finish:
						mv = cur_dev->use_time[j];
						break;
					case size_slack_ratio:
						mv =  tasks[i].length / (tasks[i].deadline - cur_dev->use_time[j]);
						break;
					case energy_finish:
						mv = add_energy + cur_dev->use_time[j];
						break;
					default:
						printf("To be implemented\n");
						exit(1);
				}
				if(mv < smallest)
				{
					smallest = mv;
					flag = 0;
					eft_core = j;
					eft_dev = cur_dev;
				}
			}
			cur_dev->use_time[j] -= tasks[i].length / cur_dev->cap;
		}

		
		//scheduling the task to an edge server
		for(j=0; j<num_edge; j++)
		{
			if(cur_dev->edge != j)
				continue;
			for(k=0; k<edges[j].svr_num; k++)
			{
				struct server *cur_srv = edges[j].svr + k;
				for(l=0; l<cur_srv->core; l++)
				{
					add_energy = cal_energy_dev(cur_dev, 0.0, -1);
					cur_dev->updata += tasks[i].sizeIn;
					cur_dev_energy = cal_energy_dev(cur_dev, tasks[i].sizeIn, edges[j].bw_down);
					add_energy = cur_dev_energy - add_energy;
					
					cur_srv->in_time[l] += tasks[i].sizeIn / edges[j].bw_down * 8;
					//the later of the ready time for input data and cpu
					if(cur_srv->in_time[l] > cur_srv->cpu_time[l])
						cur_srv->cpu_time[l] += cur_srv->in_time[l] - cur_srv->cpu_time[l];
					cur_srv->cpu_time[l] += tasks[i].length / cur_srv->cap;
					
					if((cur_srv->cpu_time[l] <= tasks[i].deadline) && 
						(cur_dev_energy <= cur_dev->pow_budget))
					{
						switch (metric)
						{
							case slack:
								mv = tasks[i].deadline - cur_srv->cpu_time[l];
								break;
							case finish:
								mv = cur_srv->cpu_time[l];
								break;
							case size_slack_ratio:
								mv =  tasks[i].length / (tasks[i].deadline - cur_srv->cpu_time[l]);
								break;
							case energy_finish:
								mv = add_energy + cur_srv->cpu_time[l];
								break;
							default:
								printf("To be implemented\n");
								exit(1);
						}
						if(mv < smallest)
						{
							smallest = mv;
							flag = 1;
							eft_core = l;
							eft_server = cur_srv;
							eft_edge = edges + j;
						}
					}
					cur_dev->updata -= tasks[i].sizeIn;
					
					cur_srv->in_time[l] -= tasks[i].sizeIn / edges[j].bw_down * 8;
					if(cur_srv->in_time[l] > cur_srv->cpu_time[l])
						cur_srv->cpu_time[l] -= cur_srv->in_time[l] - cur_srv->cpu_time[l];
					cur_srv->cpu_time[l] -= tasks[i].length / cur_srv->cap;
				}
			}

		}


		//scheduling the task to a cloud
		for(j=0; j<num_cloud; j++)
		{
			struct cloud *cur_cloud = clouds + j;
			struct vm *cur_vm = cur_cloud->vms;
			while(cur_vm)
			{
				add_energy = cal_energy_dev(cur_dev, 0.0, -1);
				cur_dev->updata += tasks[i].sizeIn;
				cur_dev_energy = cal_energy_dev(cur_dev, tasks[i].sizeIn, cur_cloud->bw_down);
				add_energy = cur_dev_energy - add_energy;
				
				cur_vm->in_time += tasks[i].sizeIn / cur_cloud->bw_down * 8;
				if(cur_vm->in_time > cur_vm->cpu_time)
					cur_vm->cpu_time += cur_vm->in_time - cur_vm->cpu_time;
				cur_vm->cpu_time += tasks[i].length / cur_cloud->cap;
				
				if((cur_vm->cpu_time <= tasks[i].deadline) && 
					(cur_dev_energy <= cur_dev->pow_budget))
				{
					switch (metric)
					{
						case slack:
							mv = tasks[i].deadline - cur_vm->cpu_time;
							break;
						case finish:
							mv = cur_vm->cpu_time;
							break;
						case size_slack_ratio:
							mv =  tasks[i].length / (tasks[i].deadline - cur_vm->cpu_time);
							break;
						case energy_finish:
							mv = add_energy + cur_vm->cpu_time;
							break;
						default:
							printf("To be implemented\n");
							exit(1);
					}
					if(mv < smallest)
					{
						smallest = mv;
						flag = 2;
						eft_vm = cur_vm;
						eft_cloud = cur_cloud;
					}
				}
				cur_dev->updata -= tasks[i].sizeIn;
				
				cur_vm->cpu_time -= tasks[i].length / cur_cloud->cap;
				if(cur_vm->in_time > cur_vm->cpu_time)
					cur_vm->cpu_time -= cur_vm->in_time - cur_vm->cpu_time;
				cur_vm->in_time -= tasks[i].sizeIn / cur_cloud->bw_down * 8;
				
				cur_vm = cur_vm->next;
			}

		}
		// rent a new VM
		if(flag < 0)
		{
			float org_energy = cal_energy_dev(cur_dev, 0.0, -1);
			cur_dev->updata += tasks[i].sizeIn;
			for(j=0; j<num_cloud; j++)
			{
				struct cloud *cur_cloud = clouds + j;
				cur_dev_energy = cal_energy_dev(cur_dev, tasks[i].sizeIn, cur_cloud->bw_down);
				add_energy = cur_dev_energy - org_energy;
				
				float ft = tasks[i].sizeIn / cur_cloud->bw_down * 8 + tasks[i].length / cur_cloud->cap;
				if((ft <= tasks[i].deadline) && (cur_dev_energy <= cur_dev->pow_budget))
				{
					switch (metric)
					{
						case slack:
							mv = tasks[i].deadline - ft;
							break;
						case finish:
							mv = ft;
							break;
						case size_slack_ratio:
							mv =  tasks[i].length / (tasks[i].deadline - ft);
							break;
						case energy_finish:
							mv = add_energy + ft;
							break;
						default:
							printf("To be implemented\n");
							exit(1);
					}
					if(mv < smallest)
					{
						smallest = mv;
						flag = 3;
						eft_cloud = cur_cloud;
					}
				}

			}

			cur_dev->updata -= tasks[i].sizeIn;
		}
		if(flag >= 0)
		{
			rst.fin_num++;
			rst.fin_size += tasks[i].length;
			rst.input += tasks[i].sizeIn;
			switch(flag)
			{
				case 0:
					rst.local_num ++;
					rst.local_size += tasks[i].length;
					rst.local_input += tasks[i].sizeIn;
					eft_dev->use_time[eft_core] += tasks[i].length / eft_dev->cap;
					break;
				case 1:
					rst.edge_num ++;
					rst.edge_size += tasks[i].length;
					rst.edge_input += tasks[i].sizeIn;

					cur_dev->updata += tasks[i].sizeIn;
					cur_dev->updata_edge[eft_edge - edges] += tasks[i].sizeIn;
					
					eft_server->in_time[eft_core] += tasks[i].sizeIn / eft_edge->bw_down * 8;
					eft_server->cpu_time[eft_core] += tasks[i].length / eft_server->cap;
					//the later of the ready time for input data and cpu
					if(eft_server->in_time[eft_core] > eft_server->cpu_time[eft_core])
						eft_server->cpu_time[eft_core] += eft_server->in_time[eft_core] - eft_server->cpu_time[eft_core];
					break;
				case 2:
					rst.cloud_num ++;
					rst.cloud_size += tasks[i].length;
					rst.cloud_input += tasks[i].sizeIn;
					cur_dev->updata_cloud[eft_cloud - clouds] += tasks[i].sizeIn;
					eft_vm->in_time += tasks[i].sizeIn / eft_cloud->bw_down * 8;
					if(eft_vm->in_time > eft_vm->cpu_time)
						eft_vm->cpu_time += eft_vm->in_time - eft_vm->cpu_time;
					eft_vm->cpu_time += tasks[i].length / eft_cloud->cap;
					break;
				case 3:
					rst.cloud_num ++;
					rst.cloud_size += tasks[i].length;
					rst.cloud_input += tasks[i].sizeIn;
					cur_dev->updata_cloud[eft_cloud - clouds] += tasks[i].sizeIn;
					eft_vm = (struct vm *)malloc(sizeof(struct vm));
					memset(eft_vm, 0, sizeof(struct vm));
					eft_vm->cloud = eft_cloud;
					eft_vm->in_time = tasks[i].sizeIn / eft_cloud->bw_down * 8;
					eft_vm->cpu_time = eft_vm->in_time + tasks[i].length / eft_cloud->cap;
					eft_vm->next = eft_cloud->vms;
					eft_cloud->vms = eft_vm;
					break;
				default:
					printf("No such situation: flag = %d\n", flag);
			}
		}
	}
	float max_time = 0.0;
	for(i=0; i<num_dev; i++)
	{
		rst.dev_energy += cal_energy_dev(devs + i, 0.0, 0);
		max_time = find_max_time(devs[i].use_time, devs[i].core);
		rst.local_used += max_time * devs[i].cap * devs[i].core;
		if(rst.ft < max_time)
			rst.ft = max_time;
	}
	for(i=0; i<num_edge; i++)
	{
		for(j=0; j<edges[i].svr_num; j++)
		{
			struct server *cur_srv = edges[i].svr + j;
			max_time = find_max_time(cur_srv->cpu_time, cur_srv->core);
			rst.edge_used += max_time * cur_srv->cap * cur_srv->core;
			if(rst.ft < max_time)
				rst.ft = max_time;
		}
	}
	for(i=0; i<num_cloud; i++)
	{
		rst.cost += cal_cost(clouds + i);
	}
	rst.total_used = rst.local_used + rst.edge_used + rst.cloud_used;
	return rst.fin_num;
}


void end();
